package geniusweb.voting.votingevaluators;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;

import geniusweb.actions.PartyId;
import geniusweb.inform.Agreements;
import geniusweb.issuevalue.Bid;
import geniusweb.voting.CollectedVotes;
import geniusweb.voting.VotingEvaluator;

/**
 * Take the first possible agreement, and take the largest possible agreement.
 * Stop after first agreement.
 */
public class LargestAgreement implements VotingEvaluator {
	@JsonIgnore
	private final CollectedVotes allVotes;

	// caching
	@JsonIgnore
	private Agreements maxAgreements = null;

	@JsonCreator
	public LargestAgreement() {
		this(new CollectedVotes(Collections.emptyMap(),
				Collections.emptyMap()));
	}

	protected LargestAgreement(CollectedVotes votes) {
		this.allVotes = votes;
	}

	@Override
	public Agreements getAgreements() {
		if (maxAgreements == null)
			maxAgreements = collectVotes();
		return maxAgreements;
	}

	@Override
	public boolean isFinished() {
		HashSet<PartyId> parties = new HashSet<PartyId>(
				allVotes.getVotes().keySet());
		parties.removeAll(getAgreements().getMap().keySet());
		return parties.size() < 2;
	}

	@Override
	public VotingEvaluator create(CollectedVotes votes) {
		return new LargestAgreement(votes);
	}

	/**
	 * @return the agreements that were reached from the given votes, using the
	 *         greedy algorithm picking the largets votesets.
	 */
	protected Agreements collectVotes() {
		Agreements agreement = new Agreements();

		Map<Bid, Set<PartyId>> agrees = allVotes.getMaxAgreements();
		if (agrees.isEmpty())
			return agreement;// none

		// find the one with max group power
		Bid maxbid = agrees.keySet().stream().max(Comparator
				.comparingInt(bid -> allVotes.getTotalPower(agrees.get(bid))))
				.get();
		agreement = agreement.with(new Agreements(maxbid, agrees.get(maxbid)));
		return agreement;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((allVotes == null) ? 0 : allVotes.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		LargestAgreement other = (LargestAgreement) obj;
		if (allVotes == null) {
			if (other.allVotes != null)
				return false;
		} else if (!allVotes.equals(other.allVotes))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "LargestAgreement";
	}
}
